@@ -1,10 +1,10 @@
-/* $Id: logrel.cpp 89632 2013-10-07 21:19:37Z bird $ */
+/* $Id: logrel.cpp 118412 2017-10-17 14:26:02Z bird $ */
 /** @file
  * Runtime VBox - Release Logger.
  */
 
 /*
- * Copyright (C) 2006-2012 Oracle Corporation
+ * Copyright (C) 2006-2017 Oracle Corporation
  *
  * This file is part of VirtualBox Open Source Edition (OSE), as
  * available from http://www.virtualbox.org. This file is free software;
@@ -25,9 +25,9 @@
  */
 
 
-/*******************************************************************************
-*   Header Files                                                               *
-*******************************************************************************/
+/*********************************************************************************************************************************
+*   Header Files                                                                                                                 *
+*********************************************************************************************************************************/
 #include <iprt/log.h>
 #include "internal/iprt.h"
 
@@ -57,9 +57,9 @@
 #endif
 
 
-/*******************************************************************************
-*   Global Variables                                                           *
-*******************************************************************************/
+/*********************************************************************************************************************************
+*   Global Variables                                                                                                             *
+*********************************************************************************************************************************/
 #ifdef IN_RC
 /** Default release logger instance. */
 extern "C" DECLIMPORT(RTLOGGERRC)   g_RelLogger;
@@ -69,13 +69,7 @@ static PRTLOGGER                    g_pRelLogger;
 #endif /* !IN_RC */
 
 
-/**
- * Gets the default release logger instance.
- *
- * @returns Pointer to default release logger instance.
- * @returns NULL if no default release logger instance available.
- */
-RTDECL(PRTLOGGER)   RTLogRelDefaultInstance(void)
+RTDECL(PRTLOGGER)   RTLogRelGetDefaultInstance(void)
 {
 #ifdef IN_RC
     return &g_RelLogger;
@@ -83,7 +77,33 @@ RTDECL(PRTLOGGER)   RTLogRelDefaultInstance(void)
     return g_pRelLogger;
 #endif /* !IN_RC */
 }
-RT_EXPORT_SYMBOL(RTLogRelDefaultInstance);
+RT_EXPORT_SYMBOL(RTLogRelGetDefaultInstance);
+
+
+RTDECL(PRTLOGGER)   RTLogRelGetDefaultInstanceEx(uint32_t fFlagsAndGroup)
+{
+#ifdef IN_RC
+    PRTLOGGER pLogger = &g_RelLogger;
+#else /* !IN_RC */
+    PRTLOGGER pLogger = g_pRelLogger;
+#endif /* !IN_RC */
+    if (pLogger)
+    {
+        if (pLogger->fFlags & RTLOGFLAGS_DISABLED)
+            pLogger = NULL;
+        else
+        {
+            uint16_t const fFlags = RT_LO_U16(fFlagsAndGroup);
+            uint16_t const iGroup = RT_HI_U16(fFlagsAndGroup);
+            if (   iGroup != UINT16_MAX
+                 && (   (pLogger->afGroups[iGroup < pLogger->cGroups ? iGroup : 0] & (fFlags | (uint32_t)RTLOGGRPFLAGS_ENABLED))
+                     != (fFlags | (uint32_t)RTLOGGRPFLAGS_ENABLED)))
+            pLogger = NULL;
+        }
+    }
+    return pLogger;
+}
+RT_EXPORT_SYMBOL(RTLogRelGetDefaultInstanceEx);
 
 
 #ifndef IN_RC
@@ -122,7 +142,7 @@ RTDECL(void) RTLogRelLoggerV(PRTLOGGER pLogger, unsigned fFlags, unsigned iGroup
      */
     if (!pLogger)
     {
-        pLogger = RTLogRelDefaultInstance();
+        pLogger = RTLogRelGetDefaultInstance();
         if (!pLogger)
             return;
     }
@@ -156,7 +176,7 @@ RT_EXPORT_SYMBOL(RTLogRelPrintfV);
  */
 RTDECL(bool) RTLogRelSetBuffering(bool fBuffered)
 {
-    PRTLOGGER pLogger = RTLogRelDefaultInstance();
+    PRTLOGGER pLogger = RTLogRelGetDefaultInstance();
     if (pLogger)
         return RTLogSetBuffering(pLogger, fBuffered);
     return false;
